
#if defined(CONFIG_CLOCK_CONTROL_NRF2)
#include <hal/nrf_lrcconf.h>
#endif

#include <nrfx_spim.h>
#include <string.h>
#include <stdio.h>
#include <zephyr/drivers/gpio.h>
#include "encoder.h"
#include "support.h"

void event_handler_spi(nrfx_spim_evt_t const * p_event,void *p_context){
}

int encoder_init(const encoder_cfg_t *cfg,encoder_t* inst){ 

    //Initialisation of SPIM
    nrfx_spim_config_t spi_config = NRFX_SPIM_DEFAULT_CONFIG(
        cfg->pin_sclk,
        cfg->pin_mosi,
        cfg->pin_miso,
        NRF_SPIM_PIN_NOT_CONNECTED
    );
    spi_config.frequency = cfg->freq;

    int err = nrfx_spim_init(&inst->spim_instance, &spi_config, event_handler_spi, NULL);
    if (err == NRFX_SUCCESS) {
        printk("SPI initialized successfully.\n");
    }
    else{
        printk("Error while SPI initialized, Error code %d \n",err);
        return err;
    }

    inst->pin_cs = cfg->pin_cs;

    inst->reg_spi_end = (uint32_t*) nrfx_spim_end_event_address_get(&inst->spim_instance);

    // Initialisation of GPIO PIN for CS
    err = gpio_pin_configure(inst->gpio_dev, cfg->pin_cs,GPIO_OUTPUT_ACTIVE);
    if (err){
        //printk("Error while GPIO initialized, Error code %d \n",err);
        return err;
    }

    //Reset Encoder 
    err = encoder_reset(inst);
    if (err){
        //printk("Error while resetting Encoder, Error code %d \n",err);
        return err;
    }

    return 0;

}


int encoder_cmd(encoder_t* inst, const uint8_t cmd[2], uint8_t resp[2]){

    // Setting CS to LOW for T_Clk (See datasheet)
    gpio_pin_set(inst->gpio_dev, inst->pin_cs, 0);
    k_busy_wait(3);

    // Transfer first byte
    nrfx_spim_xfer_desc_t xfer_desc = NRFX_SPIM_XFER_TRX(
        cmd,  1,
        resp, 1
    );

    // Setting register to zero
    
    *inst->reg_spi_end = 0;


    int err = nrfx_spim_xfer(&inst->spim_instance, &xfer_desc, NRFX_SPIM_FLAG_NO_XFER_EVT_HANDLER);
    if (err != NRFX_SUCCESS) {
        printk("SPI transfer failed: %d\n", err);
        gpio_pin_set(inst->gpio_dev, inst->pin_cs, 1);
        set_error(); 
        return err;
    }

    // Watting for transfer to be finished
    while (*inst->reg_spi_end == 0) {}; 
    *inst->reg_spi_end = 0;
    
    //Waiting for T_B (See datasheet)
    k_busy_wait(3);

    // Transfer second byte
    nrfx_spim_xfer_desc_t xfer_desc_2 = NRFX_SPIM_XFER_TRX(
        cmd+1,  1,
        resp+1, 1
    );

    err = nrfx_spim_xfer(&inst->spim_instance, &xfer_desc_2, NRFX_SPIM_FLAG_NO_XFER_EVT_HANDLER);
    if (err != NRFX_SUCCESS) {
        printk("SPI transfer failed: %d\n", err);
        set_error();
        gpio_pin_set(inst->gpio_dev, inst->pin_cs, 1);
        return err;
    }

    //Waitting for second Tranfer to be finished
    while (*inst->reg_spi_end == 0) {}; 
    *inst->reg_spi_end = 0;
    

    //Releasing CS after T_R (See datascheet)
    k_busy_wait(4);
    gpio_pin_set(inst->gpio_dev, inst->pin_cs, 1);
    
    return 0;
}

int encoder_reset(encoder_t* inst){
    const uint8_t reset[] = {0x00, 0x60};
    uint8_t resp[] = {0,0};

    return encoder_cmd(inst, reset, resp);
}

int encoder_read_pos(encoder_t* inst,uint16_t *pos){
    const uint8_t read[] = {0x00, 0x00};
    uint8_t resp[] = {0,0};

    int err = encoder_cmd(inst, read, resp);
    if(err)return err;

    //Calculating position
    uint16_t word = (uint16_t)resp[0] << 8 | resp[1]; 
    *pos = word & 0x3FFF; // lower 14 bits

    //Parse checkbits
    bool k1 = (resp[0] & (1<<7)) >> 7;
    bool k0 = (resp[0] & (1<<6)) >> 6;

    //compute checkbits
    
    bool k1_check = (word & 0b10)  >> 1;
    for(int i = 3; i<= 13; i += 2){
        k1_check ^= (word & (1 << i))  >> i;
    }
    k1_check = !k1_check;

    bool k0_check = (word & 0b01);
    for(int j = 2; j <= 14; j += 2){
        k0_check ^= (word &(1 << j)) >> j;
    }
    k0_check = !k0_check;

    if(k0_check != k0 || k1_check != k1){
        set_error();
        printk("Check bits are not correct");
        return (k1 << 3) | (k0 << 2) | (k1_check << 1) | k0_check;
    } 
    return 0;
}