/*
 * Copyright (c) 2018 Nordic Semiconductor ASA
 *
 * SPDX-License-Identifier: LicenseRef-Nordic-5-Clause
 */
#include <stdio.h>
#include <zephyr/device.h>
#include <zephyr/devicetree.h>
#include <zephyr/drivers/clock_control.h>
#include <zephyr/drivers/gpio.h>
#include <zephyr/logging/log.h>
#include <esb.h>
#include <zephyr/kernel.h>
#include <zephyr/ipc/ipc_service.h>
#include <stdatomic.h>

#include "esb_net.h"
#include "common.h"

#define STACK_SIZE 1024
#define PRIORITY   5

K_THREAD_STACK_DEFINE(ipc_thread_stack, STACK_SIZE);

K_SEM_DEFINE(received_package,1,1);
K_SEM_DEFINE(ipc_request,0,1);

payload_t payload = {};

void com_event_handler(struct esb_evt const *event)
{
    if (event->evt_id == ESB_EVENT_RX_RECEIVED) {
		int err = k_sem_take(&received_package,K_NO_WAIT);
		if(err){
			return;
		}
		struct esb_payload data_rx;
		esb_read_rx_payload(&data_rx);
		payload_read(data_rx.data,&payload);
		k_sem_give(&received_package);
	}
}

static void ep_received(const void *data, size_t len, void *priv)
{
	if (priv) {
        k_sem_give((struct k_sem *)priv);
    }
}

void ipc_thread_envent(void *arg1, void *arg2, void *arg3)
{
    while (true) {
		k_sem_take(&ipc_request,K_FOREVER);
		k_sem_take(&received_package,K_FOREVER);
		int ret = ipc_service_send(arg1, &payload, sizeof(payload));
		if (ret < 0) {
			printk("IPC send failed with ret %d\n", ret);
		}
		k_sem_give(&received_package);
    }
}

K_THREAD_STACK_DEFINE(ipc_stack,STACK_SIZE);

int main(void)
{
	int err;
	
	err = clocks_start();
	if (err) {
		return 0;
	}

	err = esb_initialize(com_event_handler);
	if (err) {
		printk("ESB initialization failed, err %d", err);
		return 0;
	}

	struct ipc_ept ep;

	const struct device *ipc0_instance = DEVICE_DT_GET(DT_NODELABEL(ipc0));

	struct ipc_ept_cfg ep_cfg = {
		.name = "ep0",
		.cb = {
			.received = ep_received,
		},
		.priv = &ipc_request,
	};
	
	//Initialisation of Interprocessor Comunication
	err = ipc_init(ipc0_instance,&ep_cfg,&ep);
	if(err) {
		printk("IPC Initialisation failed, Error Code: %d",err);
		return 0;
	}

	//Creating thread for Interprocessor Comunication
	struct k_thread *ipc_thread = k_malloc(sizeof(struct k_thread));
	if (!ipc_thread) {
		printk("Failed to allocate memory for ipc_thread\n");
		return 0;
	}
	memset(ipc_thread, 0, sizeof(struct k_thread));
	k_thread_create(ipc_thread, ipc_stack, K_THREAD_STACK_SIZEOF(ipc_stack), 
                ipc_thread_envent, &ep, ep_cfg.priv, NULL, 5, 0, K_NO_WAIT);


	//Starting to recive ESB
	err = esb_start_rx();
	if (err) {
		printk("RX setup failed, err %d", err);
		return 0;
	}

	printk("Initialization complete");

	while (true)
	{
		k_sleep(K_FOREVER);
	}
	
	return 0;
}
